// CirclePoints.cc -- functions to calculate points on a circle.
// @author Eric Laroche
// @version @(#)$Id: CirclePoints.cc,v 1.4 1998/05/20 10:13:54 laroche Exp $
//
// CirclePoints.cc -- functions to calculate points on a circle.
// Copyright (C) 1998 Eric Laroche.
//
// This program is free software;
// you can redistribute it and/or modify it.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.


#include "CirclePoints.h"


// As the regular polygons (that make up the points of the circle)
// are (depending on the number of edges) somewhat symmetric, not all
// points need to be calculated.
// Some of the symmetries allow easy transformations of symmetrical
// points.

// Symmetry cases considered:
// (2n+1) edges, mirror: (1 0); 1 additional point
// 2(2n+1) edges, mirrors: (1 0), (0 1); 2 additional points
// 4(2n+1) edges, mirrors: (1 0), (0 1), (1 1), (1 -1); 4 additional points
// 8(2n+1) edges, mirrors: (1 0), (0 1), (1 1), (1 -1); 8 additional points


#include <math.h>


static void calculatepoints(CirclePoint* points, int edges); // calculate the points
static int npoints(int edges); // number of points needed


static double const peri = 6.28318530718; // 2pi
static double const diag = 0.707106781187; // sqrt(2)/2


CirclePoints::CirclePoints(int n)
{
	this->edges = n;
	this->points = new CirclePoint[npoints(n)];
	calculatepoints(points, n);
	this->cache = new CirclePoint[n];
	for ( int i = 0; i < n; i++ ) // init the stuff
		this->cache[i].x = this->cache[i].y = 0;
	cachescale = 0;
}


CirclePoints::~CirclePoints()
{
	delete[] this->points;
	delete[] this->cache;
}


// fill buf with points
void CirclePoints::getpoints(CirclePoint* buf, double scale)
{
	if ( cachescale != scale ) // check if cached
	{
		cachescale = scale;
		CirclePoint *p = this->points, *c = this->cache; // to make the code less bloated

		int n = npoints(this->edges);
		int i = 0, j;

		// fill the first half/quater/eigth
		// (1 0)
		c[i].x = 1, c[i].y = 0, i++;
		// (1 0 0 1)
		for ( j = 0; j < n; j++ )
			c[i].x = p[j].x, c[i].y = p[j].y, i++;
		// adjust for scale
		if ( scale != 1 )
			for ( j = 0; j <= n; j++ )
				c[j].x *= scale, c[j].y *= scale;

		// copy to the rest
		if ( this->edges % 2 != 0 )
		{
			// (1 0 0 -1) reversed
			for ( j = n; j > 0; j-- )
				c[i].x = c[j].x, c[i].y = - c[j].y, i++;
		}
		else if ( this->edges % 4 != 0 )
		{
			// (-1 0 0 1) reversed
			for ( j = n; j > 0; j-- )
				c[i].x = - c[j].x, c[i].y = c[j].y, i++;
			// (-1 0)
			c[i].x = - c[0].x, c[i].y = 0, i++;
			// (-1 0 0 -1)
			for ( j = 1; j <= n; j++ )
				c[i].x = - c[j].x, c[i].y = - c[j].y, i++;
			// (1 0 0 -1) reversed
			for ( j = n; j > 0; j-- )
				c[i].x = c[j].x, c[i].y = - c[j].y, i++;
		}
		else if ( this->edges % 8 != 0 )
		{
			// (0 1 1 0) reversed
			for ( j = n; j > 0; j-- )
				c[i].x = c[j].y, c[i].y = c[j].x, i++;
			// (0 1)
			c[i].x = 0, c[i].y = c[0].x, i++;
			// (0 -1 1 0)
			for ( j = 1; j <= n; j++ )
				c[i].x = - c[j].y, c[i].y = c[j].x, i++;
			// (-1 0 0 1) reversed
			for ( j = n; j > 0; j-- )
				c[i].x = - c[j].x, c[i].y = c[j].y, i++;
			// (-1 0)
			c[i].x = - c[0].x, c[i].y = 0, i++;
			// (-1 0 0 -1)
			for ( j = 1; j <= n; j++ )
				c[i].x = - c[j].x, c[i].y = - c[j].y, i++;
			// (0 -1 -1 0) reversed
			for ( j = n; j > 0; j-- )
				c[i].x = - c[j].y, c[i].y = - c[j].x, i++;
			// (0 -1)
			c[i].x = 0, c[i].y = - c[0].x, i++;
			// (0 1 -1 0)
			for ( j = 1; j <= n; j++ )
				c[i].x = c[j].y, c[i].y = - c[j].x, i++;
			// (1 0 0 -1) reversed
			for ( j = n; j > 0; j-- )
				c[i].x = c[j].x, c[i].y = - c[j].y, i++;
		}
		else
		{
			// diag(1 1)
			c[i].x = c[i].y = diag * c[0].x, i++;
			// (0 1 1 0) reversed
			for ( j = n; j > 0; j-- )
				c[i].x = c[j].y, c[i].y = c[j].x, i++;
			// (0 1)
			c[i].x = 0, c[i].y = c[0].x, i++;
			// (0 -1 1 0)
			for ( j = 1; j <= n; j++ )
				c[i].x = - c[j].y, c[i].y = c[j].x, i++;
			// diag(-1 1)
			c[i].x = - (c[i].y = diag * c[0].x), i++;
			// (-1 0 0 1) reversed
			for ( j = n; j > 0; j-- )
				c[i].x = - c[j].x, c[i].y = c[j].y, i++;
			// (-1 0)
			c[i].x = - c[0].x, c[i].y = 0, i++;
			// (-1 0 0 -1)
			for ( j = 1; j <= n; j++ )
				c[i].x = - c[j].x, c[i].y = - c[j].y, i++;
			// diag(-1 -1)
			c[i].x = c[i].y = - diag * c[0].x, i++;
			// (0 -1 -1 0) reversed
			for ( j = n; j > 0; j-- )
				c[i].x = - c[j].y, c[i].y = - c[j].x, i++;
			// (0 -1)
			c[i].x = 0, c[i].y = - c[0].x, i++;
			// (0 1 -1 0)
			for ( j = 1; j <= n; j++ )
				c[i].x = c[j].y, c[i].y = - c[j].x, i++;
			// diag(1 -1)
			c[i].y = - (c[i].x = diag * c[0].x), i++;
			// (1 0 0 -1) reversed
			for ( j = n; j > 0; j-- )
				c[i].x = c[j].x, c[i].y = - c[j].y, i++;
		}
	}

	for ( int i = 0; i < this->edges; i++ )
		buf[i].x = this->cache[i].x, buf[i].y = this->cache[i].y;
}


// number of points needed
static int npoints(int edges)
{
	if ( edges % 2 != 0 )
		return edges / 2; //(2n+1)
	if ( edges % 4 != 0 )
		return edges / 4; //2(2n+1)
	if ( edges % 8 != 0 )
		return edges / 8; //4(2n+1)
	// no more symmetries considered
	return edges / 8 - 1; //8n
}


// hint for optimal number of points
int CirclePoints::hint(int n)
{
	// round up to next multiple of 8
	n += (8 - 1);
	n -= n % 8;
	return n;
}


// calculate the points
static void calculatepoints(CirclePoint* points, int edges)
{
	double phi0 = peri / edges, phi;
	int n = npoints(edges), i;

	for ( i = 0, phi = phi0; i < n; i++, phi += phi0 )
		points[i].x = cos(phi), points[i].y = sin(phi);
}
