// EpGlassProfile.cc -- functions to describe the profile of a glass.
// @author Eric Laroche
// @version @(#)$Id: EpGlassProfile.cc,v 1.1 1998/05/19 20:37:03 laroche Exp $
//
// EpGlassProfile.cc -- functions to describe the profile of a glass.
// Copyright (C) 1998 Eric Laroche.
//
// This program is free software;
// you can redistribute it and/or modify it.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.


#include "EpGlassProfile.h"


#include <math.h>

static double max2(double a, double b) {return a > b ? a : b;}


EpGlassProfile::EpGlassProfile(
	double FootRadius,
	double MinimalStemRadius,
	double StemLength,
	double MaximalCupRadius,
	double LowerCupLength,
	double OpeningRadius,
	double UpperCupLength) :
	FootRadius(FootRadius),
	MinimalStemRadius(MinimalStemRadius),
	StemLength(StemLength),
	MaximalCupRadius(MaximalCupRadius),
	LowerCupLength(LowerCupLength),
	OpeningRadius(OpeningRadius),
	UpperCupLength(UpperCupLength)
{
}


// glass length
double EpGlassProfile::length()
{
	return StemLength + LowerCupLength + UpperCupLength;
}


// maximal glass radius
double EpGlassProfile::radius()
{
	return max2(FootRadius, max2(MinimalStemRadius, max2(MaximalCupRadius, OpeningRadius)));
}


// stem: negative exponential, approximate stemradius
double EpGlassProfile::Stem(double z)
{
	return (FootRadius - MinimalStemRadius) * exp(-25 * (z / StemLength)) + MinimalStemRadius;
}


// lower part: quarter of an ellipsoid: horizontal at lowest, vertical at LowerCupLength
double EpGlassProfile::LowerCup(double z)
{
	// get some symbolic math to solve (though quite easy): CForm[r /.
	// Solve[(zz/LowerCupLength)^2+(r/MaximalCupRadius)^2==1,r][[-1,1]] /.
	// zz->LowerCupLength-z] ...and get rid of the pow() calls
	return sqrt(MaximalCupRadius * MaximalCupRadius - (MaximalCupRadius * MaximalCupRadius *
		(LowerCupLength - z) * (LowerCupLength - z)) / (LowerCupLength * LowerCupLength));
}


// upper part: paraboloid, vertical at 0
double EpGlassProfile::UpperCup(double z)
{
	// CForm[p z^2 + MaximalCupRadius /. Solve[(p z^2 + MaximalCupRadius == r) /.
	// {z->UpperCupLength, r->OpeningRadius}, p][[1, 1]]]
	return MaximalCupRadius - ((MaximalCupRadius - OpeningRadius) * z * z) /
		(UpperCupLength * UpperCupLength);
}


// profile (i.e. radius depending on the 'z coordinate)
double EpGlassProfile::profile(double z)
{
	if ( z < 0 )
		return 0;
	if ( z < StemLength )
		return Stem(z);
	if ( z < StemLength + LowerCupLength )
		return max2(Stem(z), LowerCup(z - StemLength));
	if ( z <= StemLength + LowerCupLength + UpperCupLength )
		return UpperCup(z - (StemLength + LowerCupLength));
	return 0;
}
