// VrmlGlass.cc -- generate output for a VRML 1.0 representation of a glass.
// @author Eric Laroche
// @version @(#)$Id: VrmlGlass.cc,v 1.2 1998/05/20 20:24:33 laroche Exp $
//
// VrmlGlass.cc -- generate output for a VRML 1.0 representation of a glass.
// Copyright (C) 1998 Eric Laroche.
//
// This program is free software;
// you can redistribute it and/or modify it.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.


#include "VrmlGlass.h"

#include "CirclePoints.h"
#include "GlassProfile.h"


#include <iostream.h>
#include <strstream.h>
#include <stdlib.h>
#include <string.h>


static char* floatform(char* buf, int size, double x);
static void glass(ostream& out, char const* indent, GlassProfile* gp, int zresolution, int phiresolution);


// VRML document
void vrmlglass(ostream& out, GlassProfile* gp)
{
	vrmlglass(out, gp,
		// VRML polygon resolution
		40,
		CirclePoints::hint(32)); // best when a multiple of 8
}


// VRML document
void vrmlglass(ostream& out, GlassProfile* gp, int zresolution, int phiresolution)
{
	// VRML 1.0 header
	out << "" << "#VRML V1.0 ascii" << "\n";

	// VRML root object
	out << "" << "Separator" << "\n";
	out << "" << "{" << "\n";

	// turn 'z coordinate to the vertical
	out << " " << "MatrixTransform" << "\n";
	out << " " << "{" << "\n";
	out << "  " << "matrix" << "\n";
	out << "   " << "1 0 0 0" << "\n";
	out << "   " << "0 0 -1 0" << "\n";
	out << "   " << "0 1 0 0" << "\n";
	out << "   " << "0 0 0 1" << "\n";
	out << " " << "}" << "\n";

	// white background
	out << " " << "DEF BackgroundColor Info {string \"1 1 1\"}" << "\n";

	// glass
	glass(out, " ", gp, zresolution, phiresolution);

	// end root object
	out << "" << "}" << "\n";
}


// glass
static void glass(ostream& out, char const* indent, GlassProfile* gp, int zresolution, int phiresolution)
{
	// glass
	out << indent << "" << "Separator" << "\n";
	out << indent << "" << "{" << "\n";

	CirclePoints* cp = new CirclePoints(phiresolution);
	CirclePoint* buf = new CirclePoint[phiresolution];

	// points
	out << indent << " " << "Coordinate3" << "\n";
	out << indent << " " << "{" << "\n";
	out << indent << "  " << "point" << "\n";
	out << indent << "  " << "[" << "\n";

	int nz;
	for ( nz = 0; nz <= zresolution;  nz++ )
	{
		double z = gp->length() / zresolution * nz;
		if ( nz == 0 ) // try to be precise at the ends
			z = 0;
		else if ( nz == zresolution )
			z = gp->length();
		cp->getpoints(buf, gp->profile(z));

		static int const formbufsize = 32;
		char formbufx[formbufsize], formbufy[formbufsize], formbufz[formbufsize];

		for ( int nphi = 0; nphi < phiresolution; nphi++ )
			out
				// don't indent to save space
				// << indent << "   "
				<< floatform(formbufx, sizeof ( formbufx ), buf[nphi].x)
				<< " " << floatform(formbufy, sizeof ( formbufy ), buf[nphi].y)
				<< " " << floatform(formbufz, sizeof ( formbufz ), z)
				<< "," << "\n";
	}

	// points end
	out << indent << "  " << "]" << "\n";
	out << indent << " " << "}" << "\n";

	// polygons
	out << indent << " " << "IndexedFaceSet" << "\n";
	out << indent << " " << "{" << "\n";
	out << indent << "  " << "coordIndex" << "\n";
	out << indent << "  " << "[" << "\n";

	for ( nz = 0; nz < zresolution;  nz++ )
	{
		for ( int nphi = 0; nphi < phiresolution; nphi++ )
		{
			int i = nphi, j = i + 1;
			if ( j == phiresolution )
				j = 0;
			i += nz * phiresolution;
			j += nz * phiresolution;

			out
				// don't indent to save space
				// << indent << "   "
				<< i
				<< ", " << j
				<< ", " << (j + phiresolution)
				<< ", " << (i + phiresolution)
				<< ", " << "-1"
				<< "," << "\n";
		}
	}

	// polygons end
	out << indent << "  " << "]" << "\n";
	out << indent << " " << "}" << "\n";

	delete buf;
	delete cp;

	// glass end
	out << indent << "" << "}" << "\n";
}


// string representation of the floating point values
static char* floatform(char* buf, int size, double x)
{
	ostrstream s(buf, size);

	// fixed, five digits
	s.setf(ios::fixed, ios::floatfield);
	s.precision(5);
	s << x << ends << flush;

	// cut trailing zeros
	if ( strchr(buf, '.') )
	{
		char* p = &buf[strlen(buf) - 1];
		while ( *p == '0' )
			*p-- = '\0';
		if ( *p == '.' )
			*p-- = '\0';
	}

	return buf;
}
